/*
 *  Copyright (C) 2002-2005  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <assert.h>
#include <list>

#include "dosbox.h"
#include "inout.h"
#include "mixer.h"
#include "pic.h"
#include "setup.h"
#include "timer.h"
#include "math.h"
#include "regs.h"
#include "serialport.h"

#define SERIALBASERATE 115200
#define LOG_UART LOG_MSG

CSerialList seriallist;


void CSerial::Timer(void)
{
	////LOG_UART("Serial port at %x: Timer", base);
	if(loopback_pending)
	{
		//LOG_UART("Serial port at %x: Loopback sent back", base);
		
		loopback_pending = false;
		receiveByte(loopback_data);
		ByteTransmitted();
	}
	//dotxint=true;
	//checkint();
}

/*void CSerial::checkint(void) {
	/* Check which irq to activate */
	//TODO Check for line status changes, but we can't get errors anyway
	//Since we only check once every millisecond, we just ignore the fifosize parameter
	/*if ((ier & 0x1) && rqueue->inuse()) {
//		LOG_MSG("RX IRQ with %d",rqueue->inuse());
		iir=0x4;
	} else if ((ier & 0x2) && !tqueue->inuse() && dotxint) {
		iir=0x2;
	} else if ((ier & 0x8) && (mstatus & 0x0f)) {
		iir=0x0;
	} else {
		iir=0x1;
		PIC_DeActivateIRQ(irq);
			return;
		}
	if (mctrl & 0x8) PIC_ActivateIRQ(irq);
	else PIC_DeActivateIRQ(irq);
}
/**/
void SERIAL_Update(void)
{
	for (CSerial_it it=seriallist.begin();it!=seriallist.end();it++)
	{
		CSerial * serial=(*it);
		serial->Timer();
		serial->Timer2();
	}
}
/*
void CSerial::SetModemStatus(Bit8u status) {
	status&=0xf;
	Bit8u oldstatus=mstatus >> 4;
	if (oldstatus ^ status ) {
		mstatus=(mstatus & 0xf) | status << 4;
		mstatus|=(oldstatus ^ status) & ((status & 0x4) | (0xf-0x4));
	}
}*/
/*
void CSerial::write_reg(Bitu reg, Bitu val) {

	//LOG_UART("Serial port: Write to 0x%x, with 0x%x '%c'\n", reg,val,val);
	//	if (port!=9 && port!=8) LOG_MSG("Serial write %X val %x %c",port,val,val);
	switch(reg) {
	case 0x0:  // Transmit holding buffer + Divisor LSB
		if (dlab) {
			divisor_lsb = val;
			UpdateBaudrate();
			return;
		}
		if (local_loopback) rqueue->addb(val);
		else tqueue->addb(val);
		break;
	case 0x1:  // Interrupt enable register + Divisor MSB
		if (dlab) {
			divisor_msb = val;
			UpdateBaudrate();
			return;
		} else {
            ier = val;
			dotxint=true;
		}
		break;
	case 0x2:  // FIFO Control register
		FIFOenabled = (val & 0x1) > 0;
		if (val & 0x2) rqueue->clear(); //Clear receiver FIFO
		if (val & 0x4) tqueue->clear();	//Clear transmit FIFO
		if (val & 0x8) LOG(LOG_MISC,LOG_WARN)("UART:Enabled DMA mode");
		switch (val >> 6) {
		case 0:FIFOsize = 1;break;
		case 1:FIFOsize = 4;break;
		case 2:FIFOsize = 8;break;
		case 3:FIFOsize = 14;break;
		}
		break;
	case 0x3:  // Line control register
		linectrl = val;
		dlab = (val & 0x80);
		break;
	case 0xc:  // Modem control register
		mctrl=val;
		local_loopback = (val & 0x10);
		break;
	case 0x7:  // Scratch register
		scratch = val;
		break;
	default:
		////LOG_UART("Serial port: Write to 0x%x, with 0x%x '%c'\n", reg,val,val);
		break;
	}
}
*/
static void WriteSerial(Bitu port,Bitu val,Bitu iolen) {
	Bitu check=port&~0x7;
	for (CSerial_it it=seriallist.begin();it!=seriallist.end();it++){
		CSerial * serial=(*it);
		if (check==serial->base)
		{
			switch(port&0x7)
			{
			case THR_OFFSET:
				serial->Write_THR(val);
				return;
			case IER_OFFSET:
				serial->Write_IER(val);
				return;
				
			case LCR_OFFSET:
				serial->Write_LCR(val);
				return;
			case MCR_OFFSET:
				serial->Write_MCR(val);
				return;
			case MSR_OFFSET:
				serial->Write_MSR(val);
				return;
				
			case SPR_OFFSET:
				serial->Write_SPR(val);
				return;
			default:
					serial->Write_reserved(val,port&0x7);
			}
			//serial->write_reg(port&0x7,val);
		}
	}
}

static Bitu ReadSerial(Bitu port,Bitu iolen) {
	Bitu check=port&~0x7;
	
	for (CSerial_it it=seriallist.begin();it!=seriallist.end();it++){
		CSerial * serial=(*it);
		if (check==serial->base)
		{
				switch(port&0x7)
				{
				case RHR_OFFSET: return serial->Read_RHR();
				case IER_OFFSET: return serial->Read_IER();
				case ISR_OFFSET: return serial->Read_ISR();
				case LCR_OFFSET: return serial->Read_LCR();
				case MCR_OFFSET: return serial->Read_MCR();
				case LSR_OFFSET: return serial->Read_LSR();
				case MSR_OFFSET: return serial->Read_MSR();
				case SPR_OFFSET: return serial->Read_SPR();
				}
		}
	}
	return 0;
}



/*****************************************************************************/
/* Interrupt control routines												**/
/*****************************************************************************/
_inline void CSerial::rise(Bit8u priority)
{
	//LOG_UART("Serial port at %x: Rise priority 0x%x", base, priority);
	waiting_interrupts|= priority;
	WriteRealIER(IER);
}

// clears the pending interrupt, triggers other waiting interrupt
_inline void CSerial::clear(Bit8u priority)
{
	//LOG_UART("Serial port at %x: cleared priority 0x%x", base, priority);
	waiting_interrupts&= (~priority);
	WriteRealIER(IER);
}

void CSerial::WriteRealIER(Bit8u data)
{
	//LOG_UART("Serial port at %d: RealIER, value 0x%x", base, data);
	data=data&0xF;	// THE UPPER ONES ALWAYS READ 0! NOTHIN' ELSE!
	Bit8u old_pending_interrupts = pending_interrupts;
	Bit8u difference_pending_interrupts;
	
	// rise TX AGAIN when present and is being enabled
	if((data&TX_PRIORITY)&&(!(IER&TX_PRIORITY)))
		if(LSR&LSR_TX_HOLDING_EMPTY_MASK)
			waiting_interrupts|=TX_PRIORITY;

	pending_interrupts = waiting_interrupts&data;
	if((difference_pending_interrupts = (pending_interrupts ^ old_pending_interrupts)))
	{	// something in pending interrupts has changed
		if(difference_pending_interrupts & pending_interrupts)
		{	// some new bits were set
			if(!current_priority)
			{	// activate interrupt..
				if(pending_interrupts & ERROR_PRIORITY)
				{
					current_priority = ERROR_PRIORITY;
					ISR= ISR_ERROR_VAL;
				}
				else if(pending_interrupts & RX_PRIORITY)
				{
					current_priority = RX_PRIORITY;
					ISR= ISR_RX_VAL;
				}
				else if(pending_interrupts & TX_PRIORITY)
				{
					current_priority = TX_PRIORITY;
					ISR= ISR_TX_VAL;
				}
				else if(pending_interrupts & MSR_PRIORITY)
				{
					current_priority = MSR_PRIORITY;
					ISR= ISR_MSR_VAL;
				}
				//LOG_UART("Serial port at %x: Interrupt activated: priority %d", base,current_priority);
				PIC_ActivateIRQ(irq);

			} // else the new interrupts were already
			// written into pending_interrupts
		} // else no new bits were set

		if(difference_pending_interrupts&(~pending_interrupts)) // some bits were reset
		{
			if(pending_interrupts)
			{	// some more are waiting
				if(!(current_priority&pending_interrupts))
				{	// the current interrupt has been cleared
					// choose the next one
					if(pending_interrupts & ERROR_PRIORITY)
					{
						current_priority = ERROR_PRIORITY;
						ISR= ISR_ERROR_VAL;
					}
					else if(pending_interrupts & RX_PRIORITY)
					{
						current_priority = RX_PRIORITY;
						ISR= ISR_RX_VAL;
					}
					else if(pending_interrupts & TX_PRIORITY)
					{
						current_priority = TX_PRIORITY;
						ISR= ISR_TX_VAL;
					}
					else if(pending_interrupts & MSR_PRIORITY)
					{
						current_priority = MSR_PRIORITY;
						ISR= ISR_MSR_VAL;
					}
				}
			}
			else
			{	
				current_priority = NONE_PRIORITY;
				ISR= ISR_CLEAR_VAL;
				PIC_DeActivateIRQ(irq);
				//LOG_UART("Serial port at %x: Interrupt deactivated.", base);
			}
		}
	}
	IER = data;
}



/*****************************************************************************/
/* Internal register modification											**/
/*****************************************************************************/
void CSerial::changeMSR(Bit8u data)
{
	if(!(MCR&MCR_LOOPBACK_Enable_MASK))
	{
		Bit8u temp=MSR;
		// look for signal changes
		if(temp|=((data&MSR_LINE_MASK)^(MSR&MSR_LINE_MASK))>>4)
		{	// some signals changed
			temp&=MSR_delta_MASK;// clear line states
			temp|=(data&MSR_LINE_MASK);	// set new line states
			
			MSR=temp;
			rise(MSR_PRIORITY);
		}
		// else nothing happened
	}
}
void CSerial::changeMSR_Loopback(Bit8u data)
{
	Bit8u temp=MSR;
	// look for signal changes
	if(temp|=((data&MSR_LINE_MASK)^(MSR&MSR_LINE_MASK))>>4)
	{	// some signals changed
		temp&=MSR_delta_MASK;// clear line states
		temp|=(data&MSR_LINE_MASK);	// set new line states
		
		MSR=temp;
		rise(MSR_PRIORITY);
	}
	// else nothing happened
}
/*****************************************************************************/
/* A byte was received														**/
/*****************************************************************************/
	void CSerial::receiveByte(Bit8u data)
{

	if(LSR&LSR_RX_DATA_READY_MASK)	// Overrun error ;o
	{
		LOG_UART("Serial port at 0x%x: RX Overrun!", base, data);
		LSR |= LSR_OVERRUN_ERROR_MASK;
		rise(ERROR_PRIORITY);
	}
	else
	{
        RHR = data;
		LSR|= LSR_RX_DATA_READY_MASK;
		rise(RX_PRIORITY);
	}
}
/*****************************************************************************/
/* ByteSent: When a byte was sent, notify here.								**/
/*****************************************************************************/
void CSerial::ByteTransmitted()
{
	if(LSR&LSR_TX_HOLDING_EMPTY_MASK)
		{	// one space was empty
			// now both are
			LSR|=LSR_TX_EMPTY_MASK;
		}
	else	// both were full, now 1 is empty
	{
		if(MCR&MCR_LOOPBACK_Enable_MASK)
		{	// loopback mode
			LSR&=LSR_TX_HOLDING_EMPTY_MASK;
			transmitLoopbackByte(THR);
			rise(TX_PRIORITY);
		}
		else
		{	// direct "real" mode
			transmitByte(THR);
			rise(TX_PRIORITY);
		}
		
	}

}

/*****************************************************************************/
/* Transmit Holding Register, also LSB of Divisor Latch (r/w)				**/
/*****************************************************************************/
void CSerial::Write_THR(Bit8u data)
{	
	
	if((LCR & LCR_DIVISOR_Enable_MASK))
	{	
		//LOG_UART("Serial port at %x: Write to DLL value 0x%x", base, data);
		// write to DLL
		DLL = data;
		updatePortConfig(DLL, DLM, LCR);
	}
	else
	{
		//LOG_UART("Serial port at %x: Write to THR value 0x%x", base, data);
		// write to THR
		clear(TX_PRIORITY);
		
		if(LSR&LSR_TX_HOLDING_EMPTY_MASK)
		{	// one is empty	
			if(LSR&LSR_TX_EMPTY_MASK)
			{	// both are empty
				
				if(MCR&MCR_LOOPBACK_Enable_MASK)
				{	// loopback mode
					LSR&=(~LSR_TX_EMPTY_MASK);
					transmitLoopbackByte(data);
				}
				else
				{	
					// direct "real" mode
					LSR&=(~LSR_TX_EMPTY_MASK);
					
					transmitByte(data);
				}
				rise(TX_PRIORITY);
			}
			else 
			{	// only THR is empty; add byte and clear holding bit
				LSR&=(~LSR_TX_HOLDING_EMPTY_MASK);//&0x7f;
				THR = data;
			}
		}
		else
		{
			//LOG_UART("Serial port at %x: TX Overflow!", base);
			// TX overflow; how does real hardware respond to that... I do nothing
		}
	}
}


/*****************************************************************************/
/* Receive Holding Register, also LSB of Divisor Latch (r/w)				**/
/*****************************************************************************/
Bitu CSerial::Read_RHR()// Receive Holding Register, also LSB of Divisor Latch (r/w)
{	
	if((LCR & LCR_DIVISOR_Enable_MASK))
	{
		////LOG_UART("Serial port at %d: Read from DLL value 0x%x", base, DLL);
		return DLL;
	}
	else
	{
		clear(RX_PRIORITY);
		LSR&=(~LSR_RX_DATA_READY_MASK);
		//LOG_UART("Serial port at %d: Read from RHR value 0x%x", base, RHR);
		return RHR;
	}
}

/*****************************************************************************/
/* Interrupt Enable Register, also MSB of Divisor Latch (r/w)				**/
/*****************************************************************************/
// Modified by:
// - writing to it.
Bitu CSerial::Read_IER()
{
	
	if((LCR & LCR_DIVISOR_Enable_MASK))// IER or MSB?
	{
		//LOG_UART("Serial port at %d: Read from DLM value 0x%x", base, DLM);
		return DLM;
	}
	else
	{
		//LOG_UART("Serial port at %d: Read from IER value 0x%x", base, IER);
		return IER;
	}
}

void CSerial::Write_IER(Bit8u data)
{	
	if((LCR & LCR_DIVISOR_Enable_MASK))
	{	// write to DLM
		//LOG_UART("Serial port at %d: Write to DLM value 0x%x", base, data);
		DLM = data;
		updatePortConfig(DLL, DLM, LCR);
	}
	else
	{
		//LOG_UART("Serial port at %d: Write to IER value 0x%x", base, data);
		WriteRealIER(data);
		
	}
}
/*****************************************************************************/
/* Interrupt Status Register (r)											**/
/*****************************************************************************/
// modified by:
// - incoming interrupts
// - loopback mode
Bitu CSerial::Read_ISR()// Interrupt Status Register
{	
	Bit8u retval = ISR;
	//LOG_UART("Serial port at %d: Read from ISR value 0x%x", base, retval);
	
	// clear changes ISR!! mean..
	clear(TX_PRIORITY);
	return retval;
}
/*****************************************************************************/
/* Line Control Register (r/w)												**/
/*****************************************************************************/
// signal decoder configuration:
// - parity, stopbits, word length
// - send break
// - switch between RHR/THR and baud rate registers
// Modified by:
// - writing to it.
Bitu CSerial::Read_LCR()
{	
	
	//LOG_UART("Serial port at %d: Read from LCR value 0x%x", base, LCR);
	return LCR;
}
void CSerial::Write_LCR(Bit8u data)
{	
	Bit8u lcr_old = LCR;
	LCR = data;
	
	//LOG_UART("Serial port at %x: Write to LCR value 0x%x", base, data);
	
	// for debug output
	if(!(lcr_old & LCR_DIVISOR_Enable_MASK))
	{
		//if(data & LCR_DIVISOR_Enable_MASK)
			//LOG_UART("Serial port at 0x%x: Divisor-mode entered", base);
	}
	else
	{
		//if(!(data & LCR_DIVISOR_Enable_MASK))
			//LOG_UART("Serial port at 0x%x: Divisor-mode exited", base);
	}
	if((data^lcr_old)&LCR_PORTCONFIG_MASK)
	{
		//LOG_UART("Serial port at 0x%x: comm parameters changed", base);
		updatePortConfig(DLL, DLM, LCR);
	}
	
	if((LCR&LCR_BREAK_MASK)^(lcr_old&LCR_BREAK_MASK))
	toggleBreak(LCR&LCR_BREAK_MASK);

}
/*****************************************************************************/
/* Modem Control Register (r/w)												**/
/*****************************************************************************/
// Set levels of RTS and DTR, as well as loopback-mode.
// Modified by: 
// - writing to it.
Bitu CSerial::Read_MCR()
{	
	
	//LOG_UART("Serial port at %d: Read from MCR value 0x%x", base, MCR);
	return MCR;
}
void CSerial::Write_MCR(Bit8u data)
{	
	data&=0x1F;	// UPPER BITS ALWAYS 0!!!
	
	if(data==0x1f)
	{
		int x = 0;
		x++;
	}

	//LOG_UART("Serial port at %d: Write to MCR value 0x%x", base, data);
	
	if(MCR&MCR_LOOPBACK_Enable_MASK)
		if(data&MCR_LOOPBACK_Enable_MASK)
		{	// was on, is now on
			Bit8u param = 0;
			
			// RTS->CD
			// DTR->RI
			// OP1->DSR
			// OP2->CTS

			if(data&MCR_RTS_MASK) param|=MSR_CD_MASK;
			if(data&MCR_DTR_MASK) param|=MSR_RI_MASK;
			if(data&MCR_OP1_MASK) param|=MSR_DSR_MASK;
			if(data&MCR_OP2_MASK) param|=MSR_CTS_MASK;

			changeMSR_Loopback(param);

		}
		else
		{
			updateModemControlLines(data);
			// is switched off now
		}
	else
	{
		if(data&MCR_LOOPBACK_Enable_MASK)
		{	// is switched on:
			Bit8u par = 0;
			if(data&MCR_RTS_MASK) par|=MSR_CD_MASK;
			if(data&MCR_DTR_MASK) par|=MSR_RI_MASK;
			if(data&MCR_OP1_MASK) par|=MSR_DSR_MASK;
			if(data&MCR_OP2_MASK) par|=MSR_CTS_MASK;

			changeMSR(par);
			
			// go back to empty state
			LSR&=(LSR_TX_EMPTY_MASK|LSR_TX_HOLDING_EMPTY_MASK);

		}
		else
		{
			updateModemControlLines(data);
			// loopback is off
		}
	}
	MCR = data;
}

/*****************************************************************************/
/* Line Status Register (r)													**/
/*****************************************************************************/
// errors, tx registers status, rx register status
// modified by:
// - event from real serial port
// - loopback
Bitu CSerial::Read_LSR()
{
	Bit8u x = 0;
	x++;
	//x=~LSR_ERROR_MASK;
	Bitu retval = LSR;
	LSR&=(~LSR_ERROR_MASK);	// clear error bits on read
	//LOG_UART("Serial port at %d: Read from LSR value 0x%x", base, retval);
	clear(ERROR_PRIORITY);
	return retval;
}

void CSerial::Write_MSR(Bit8u val)
{
	MSR &= MSR_LINE_MASK;
	MSR |= val&MSR_delta_MASK;
}

/*****************************************************************************/
/* Modem Status Register (r)												**/
/*****************************************************************************/
// Contains status of the control input lines (CD, RI, DSR, CTS) and
// their "deltas": if level changed since last read delta = 1.
// modified by:
// - real values
// - write operation to MCR in loopback mode
Bitu CSerial::Read_MSR()
{	
	Bit8u retval = 0;
	if(!(MCR&MCR_LOOPBACK_Enable_MASK))
	{
		updateMSR();
	}
	retval = MSR;
	//LOG_UART("Serial port at %d: Read from MSR value 0x%x", base, MSR);
	clear(MSR_PRIORITY);
	// clear deltas
	MSR&=MSR_LINE_MASK;
	return retval;
}

/*****************************************************************************/
/* Scratchpad Register (r/w)												**/
/*****************************************************************************/
// Just a memory register. Not much to do here.
Bitu CSerial::Read_SPR() 
{	
	
	//LOG_UART("Serial port at %d: Read from SPR value 0x%x", base, SPR);
	return SPR;
}
void CSerial::Write_SPR(Bit8u data)
{	
	//LOG_UART("Serial port at %d: Write to SPR value 0x%x", base, data);
	SPR = data;
}
/*****************************************************************************/
/* Write_reserved															**/
/*****************************************************************************/
void CSerial::Write_reserved(Bit8u data, Bit8u address)
{	
	//LOG_UART("Serial port at %d: Write to reserved register, value 0x%x, register %x", base, data,address);
	
}

/*****************************************************************************/
/* Loopback: add byte to loopback buffer; it is received next time Timer	**/
/* function is invoked; (time needed is not emulated correctly, but I don't	**/
/* think this is sooooo important....)										**/
/*****************************************************************************/
void CSerial::transmitLoopbackByte(Bit8u val)
{
	LOG_MSG("Serial port at %x: Loopback requested", base);
	loopback_pending = true;
	loopback_data = val;

}


/*****************************************************************************/
/* Initialisation															**/
/*****************************************************************************/
void CSerial::Init_Registers()
{
	RHR = 0;	// xx
	THR = 0;	// xx
	IER = 0;
	//FCR 
	ISR = 0x1;
	LCR = 0;
	MCR = 0;
	LSR = 0x60;
	MSR = 0;

	SPR = 0xFF;

	DLL = 0x0;
	DLM = 0x0;

	pending_interrupts = 0x0;	
	current_priority = 0x0;
	waiting_interrupts = 0x0;	
		
	loopback_pending = false;

}



/*Bitu CSerial::read_reg(Bitu reg) {
	Bitu retval;
	////LOG_UART("Serial port: Write to 0x%x, with 0x%x '%c'\n", reg,val,val);
	////LOG_UART("Serial port: Read from 0x%x\n", reg);
//	if (port!=0xd && port!=0xa) LOG_MSG("REad from port %x",reg);
	switch(reg) {
	case 0x0:  //  Receive buffer + Divisor LSB
		if (dlab) {
			
			//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,divisor_lsb);
			return divisor_lsb ;
		} else {
			retval=rqueue->getb();
			//LOG_MSG("Received char %x %c",retval,retval);
			checkint();
			//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,retval);
			return retval;
		}
	case 0x1:  // Interrupt enable register + Divisor MSB
		if (dlab) {
			
			//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,divisor_msb);
			return divisor_msb ;
		} else {
			//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,ier);
			return ier;
		}
	case 0x2:  // Interrupt identification register
		retval=iir;	
		if (iir==2) {
			dotxint=false;
			iir=1;
		}
//		LOG_MSG("Read iir %d after %d",retval,iir);
		retval =retval | ((FIFOenabled) ? (3 << 6) : 0);
		//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,retval);
		return retval ;
	case 0x3:  // Line control register
		//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,linectrl);
		return linectrl;
	case 0x4:  // Modem control register
		//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,mctrl);
		return mctrl;
	case 0x5:  // Line status register
		retval = 0x40;
		if (!tqueue->inuse()) retval|=0x20;
		if (rqueue->inuse()) retval|= 1;
//		LOG_MSG("Read from line status %x",retval);
		//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,retval);
		return retval;
	case 0x6:  // modem status register 
		retval=mstatus;
		mstatus&=0xf0;
		checkint();
		//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,retval);
		return retval;
	case 0x7:  // Scratch register
		//LOG_UART("Serial port: Read from 0x%xwith 0x%x '%c'\n", reg,scratch);
		return scratch;
	default:
		//LOG_UART("Serial port: Read from reserved reg 0x%xwith 0x%x '%c'\n", reg,0);
		//LOG_DEBUG("Serial port: Read from 0x%x\n", port);
		break;
	}
	return 0x00;	
}

*/



CSerial::CSerial (Bit16u initbase, Bit8u initirq, Bit32u initbps)
{

	Bitu i;
	//Bit16u initdiv;

	Init_Registers();


	base=initbase;
	irq=initirq;
	bps=initbps;

	//local_loopback = 0;
	//ier = 0;
	//iir = 1;

	//FIFOenabled = false;	
	//FIFOsize = 1;
	//dlab = 0;
	//mstatus = 0;

	//initdiv = SERIALBASERATE / bps;
	//UpdateBaudrate();
	

	// Register I/O-Ports

	for (i=base;i<=(base+8);i++)
	{
		IO_RegisterWriteHandler(i,WriteSerial,IO_MB);
		IO_RegisterReadHandler(i,ReadSerial,IO_MB);
	}

//	rqueue=new CFifo(QUEUE_SIZE);
//	tqueue=new CFifo(QUEUE_SIZE);
};

CSerial::~CSerial(void)
{

};


void SERIAL_Init(Section* sec) {
	unsigned long args = 1;
	Section_prop * section=static_cast<Section_prop *>(sec);

//	if(!section->Get_bool("enabled")) return;

	TIMER_AddTickHandler(&SERIAL_Update);
}

